from typing import Optional, Tuple
from parser.abc.ast import AST
from parser.abc.parser_abc import ParserABC
from parser.model.ast_types_enum import ASTTypesEnum

from cc_lang.model.language_definition_classes import ExpressionCharacters, FormatCharacters, Keywords
from cpl_core.console.console import Console
from cpl_query.extension.list import List
from lexer.model.token import Token
from lexer.model.token_types import TokenTypes
from pynput.keyboard import Key
from parser.model.parser_state_enum import ParserStateEnum
from runtime.abc.runtime_service_abc import RuntimeServiceABC
from runtime.model.error import Error
from runtime.model.error_codes_enum import ErrorCodesEnum


class ParserService(ParserABC):

    def __init__(self, runtime: RuntimeServiceABC):
        self._runtime = runtime

        self._access_keywords = [
            Keywords.Public.value,
            Keywords.Private.value,
            Keywords.Static.value,
            Keywords.This.value,
        ]

    def _parse_name(self, tokens: List[Token]) -> Tuple[AST, Token]:
        """ Parses names

        Args:
            tokens (List[Token]): Tokens from lexer

        AST:
            Program
            <name>
            Program.Test
            <name> <.> <name>

        Returns:
            Tuple[AST, Token]: Name as AST and last Token of Name
        """
        name = ''
        last_token: Optional[Token] = None
        for i in range(0, tokens.count()):
            token: Token = tokens[i]

            if i == tokens.count() and token.type == TokenTypes.Format_Character and token.value == FormatCharacters.Point:
                self._runtime.error(Error(ErrorCodesEnum.Unexpected), FormatCharacters.Point.value)

            if token.type == TokenTypes.Name or token.type == TokenTypes.Format_Character and token.value == FormatCharacters.Point.value:
                name += token.value
            else:
                break

            last_token = token

        return (AST(ASTTypesEnum.Name, name, self._runtime.line_count, self._runtime.line_count), last_token)

    def _parse_args(self, tokens: List[Token]) -> Tuple[AST, Token]:
        """ Parses args

        Args:
            tokens (List[Token]): Tokens from lexer

        AST:
            ()
            <start> <end>
            (i: number)
            <start> <name> <:> <type> <end>
            (i: number, name: string)
            <start> <name> <:> <type> <,> <name> <:> <type> <end>

        Returns:
            Tuple[AST, Token]: Name as AST and last Token of Name
        """
        if tokens.where(lambda t: t.type == TokenTypes.Format_Character and t.value == FormatCharacters.Left_Parenthesis.value).count() < 1:
            self._runtime.error(Error(ErrorCodesEnum.Expected, FormatCharacters.Left_Parenthesis.value))
        else:
            i = 1

        if tokens.where(lambda t: t.type == TokenTypes.Format_Character and t.value == FormatCharacters.Right_Parenthesis.value).count() < 1:
            self._runtime.error(Error(ErrorCodesEnum.Expected, FormatCharacters.Right_Parenthesis.value))

        ast = List(AST)
        last_token: Optional[Token] = None
        expected_var = False
        # for i in range(0, tokens.count()):
        while i < tokens.count():
            token: Token = tokens[i]

            if token.type == TokenTypes.Format_Character and token.value == FormatCharacters.Right_Parenthesis.value:
                last_token = token
                break

            if expected_var and token.type != TokenTypes.Name:
                self._runtime.error(Error(ErrorCodesEnum.Unexpected, token.value))

            if token.type == TokenTypes.Name:
                if i + 1 >= tokens.count():
                    self._runtime.error(Error(ErrorCodesEnum.Expected, FormatCharacters.Colon.value))
                
                if i + 2 >= tokens.count():
                    self._runtime.error(Error(ErrorCodesEnum.Expected, Keywords.Type.value))

                var_ast = List(AST)
                var_ast.append(AST(ASTTypesEnum.Name, token.value, self._runtime.line_count, self._runtime.line_count))
                var_ast.append(AST(ASTTypesEnum.Type, tokens[i+2].value, self._runtime.line_count, self._runtime.line_count))
                ast.append(AST(ASTTypesEnum.VariableDeclaration, var_ast, self._runtime.line_count, self._runtime.line_count))
                i += 2

            if token.type == TokenTypes.Format_Character and token.value == FormatCharacters.Comma.value:
                expected_var = True

            last_token = token
            i += 1

        return (AST(ASTTypesEnum.Arguments, ast, self._runtime.line_count, self._runtime.line_count), last_token)

    def _parse_library_or_class(self, tokens: List[Token], cls=False) -> AST:
        """ Parses library or class declarations

        Args:
            tokens (List[Token]): Tokens from lexer

        AST:
            lib Main {
            <lib, class> <name> <end>
            public lib Main {
            <access> <lib, class> <name> <end>
            public lib Main {}
            <access> <lib, class> <name> <end> <end>
            public lib Main.Test {
            <access> <lib, class> <name> <.> <name> <end> <end>

        Returns:
            AST: Library or class AST
        """
        end = None
        ast = List(AST)
        i = 0
        while i < tokens.count():
            token: Token = tokens[i]

            # if line contains }
            if token.type == TokenTypes.Format_Character and token.value == FormatCharacters.Right_Brace.value:
                end = self._runtime.line_count

            elif i == tokens.count()-1 and token.type == TokenTypes.Format_Character and token.value == FormatCharacters.Left_Brace.value:
                break

            elif i == tokens.count()-1:
                self._runtime.error(Error(ErrorCodesEnum.Expected, FormatCharacters.Left_Brace.value))

            elif i == 0 and token.type == TokenTypes.Keyword and token.value in self._access_keywords:
                ast.append(AST(ASTTypesEnum.Access, token.value, self._runtime.line_count, self._runtime.line_count))

            elif i <= 1 and token.type == TokenTypes.Keyword and token.value == Keywords.Library.value and not cls:
                ast.append(AST(ASTTypesEnum.Keyword, token.value, self._runtime.line_count, self._runtime.line_count))

            elif i <= 1 and token.type == TokenTypes.Keyword and token.value == Keywords.Class.value and cls:
                ast.append(AST(ASTTypesEnum.Keyword, token.value, self._runtime.line_count, self._runtime.line_count))

            elif i >= 1 and token.type == TokenTypes.Name:
                name, last_token = self._parse_name(tokens.skip(i))
                if last_token is not None:
                    i = tokens.index(last_token)
                    ast.append(name)

            else:
                self._runtime.error(Error(ErrorCodesEnum.Unexpected, token.value))

            i += 1

        return AST(ASTTypesEnum.LibraryDeclaration if not cls else ASTTypesEnum.ClassDeclaration, ast, self._runtime.line_count, end)

    def _parse_variable(self, tokens: List[AST]) -> AST:
        """ Parses variable declarations

        Args:
            tokens (List[Token]): Tokens from lexer

        AST:
            var test: number;
            <var> <name> <type> <end>
            var test = 0;
            <var> <name> <value | int, str, bool> <end>
            var test: number = 0;
            <var> <name> <type> <value | int, str, bool> <end>
            var test: number = test;
            <var> <name> <type> <name> <end>
            var test: number = TestClass();
            <var> <name> <type> <object-assign> <end>
            private var test: number = 0;
            <access> <var> <name> <type> <end>

        Returns:
            AST: Library or class AST
        """
        end = None
        ast = List(AST)
        i = 0
        # for i in range(0, tokens.count()):
        while i < tokens.count():
            token: Token = tokens[i]

            if i == tokens.count()-1 and token.type == TokenTypes.Format_Character and token.value == FormatCharacters.Semicolon.value:
                break

            elif i == tokens.count()-1:
                self._runtime.error(Error(ErrorCodesEnum.Expected, FormatCharacters.Semicolon.value))

            elif i == 0 and token.type == TokenTypes.Keyword and token.value in self._access_keywords:
                ast.append(AST(ASTTypesEnum.Access, token.value, self._runtime.line_count, self._runtime.line_count))

            elif i >= 0 and token.type == TokenTypes.Keyword and Keywords.Variable.value:
                ast.append(AST(ASTTypesEnum.Keyword, token.value, self._runtime.line_count, self._runtime.line_count))

            elif i >= 1 and token.type == TokenTypes.Name:
                name, last_token = self._parse_name(tokens.skip(i))
                i = tokens.index(last_token)
                ast.append(name)

            elif i >= 2 and token.type == TokenTypes.Format_Character and token.value == FormatCharacters.Colon.value:
                if i+1 < tokens.count():
                    next_token = tokens[i+1] 
                    ast.append(AST(ASTTypesEnum.Type, next_token.value, self._runtime.line_count, self._runtime.line_count))
                    i += 1
                else:
                    self._runtime.error(Error(ErrorCodesEnum.Expected, ASTTypesEnum.Type.value))

            elif i >= 2 and token.type == TokenTypes.Expression_Character and token.value == ExpressionCharacters.Equal.value:
                if i+1 < tokens.count():
                    end = tokens.where(lambda t: t.type == TokenTypes.Format_Character and t.value == FormatCharacters.Semicolon.value).first_or_default()
                    value = ''
                    if end is not None:
                        for t in tokens[i+1:]:
                            t: Token = t
                            if t == end:
                                break

                            value += t.value
                        ast.append(AST(ASTTypesEnum.VariableValue, value, self._runtime.line_count, self._runtime.line_count))
                        i = tokens.index(end)
                else:
                    self._runtime.error(Error(ErrorCodesEnum.Expected, ASTTypesEnum.Type.value))

            elif i == tokens.count()-1 and token.type != TokenTypes.Format_Character and token.value != FormatCharacters.Semicolon.value:
                self._runtime.error(Error(ErrorCodesEnum.Expected, FormatCharacters.Semicolon.value))

            else:
                if i == 1 or i == 2 and token.type == TokenTypes.Type:
                    self._runtime.error(Error(ErrorCodesEnum.Unexpected, f'{token.type.name.lower()}: {token.value}'))
                else:
                    self._runtime.error(Error(ErrorCodesEnum.Unexpected, {token.value}))

            i += 1

        return AST(ASTTypesEnum.VariableDeclaration, ast, self._runtime.line_count, self._runtime.line_count)

    def _parse_func(self, tokens: List[Token]) -> AST:
        """[summary]

        Args:
            tokens (List[Token]): [description]

        Returns:
            AST: [description]
        """
        end = None
        ast = List(AST)
        i = 0
        while i < tokens.count():
            token: Token = tokens[i]

            # if line contains }
            if token.type == TokenTypes.Format_Character and token.value == FormatCharacters.Right_Brace.value:
                end = self._runtime.line_count

            elif i == tokens.count()-1 and token.type == TokenTypes.Format_Character and token.value == FormatCharacters.Left_Brace.value:
                break

            elif i == tokens.count()-1:
                self._runtime.error(Error(ErrorCodesEnum.Expected, FormatCharacters.Left_Brace.value))

            elif i == 0 and token.type == TokenTypes.Keyword and token.value in self._access_keywords:
                ast.append(AST(ASTTypesEnum.Access, token.value, self._runtime.line_count, self._runtime.line_count))

            elif i <= 1 and token.type == TokenTypes.Keyword and token.value == Keywords.Function.value:
                ast.append(AST(ASTTypesEnum.Keyword, token.value, self._runtime.line_count, self._runtime.line_count))

            elif i >= 1 and token.type == TokenTypes.Name:
                ast.append(AST(ASTTypesEnum.Name, token.value, self._runtime.line_count, self._runtime.line_count))

            elif i >= 2 and token.type == TokenTypes.Format_Character and token.value == FormatCharacters.Left_Parenthesis.value:
                args, last_token = self._parse_args(tokens.skip(i))
                if last_token is not None:
                    i = tokens.index(last_token)
                    ast.append(args)
            
            elif i >= 3 and token.type == TokenTypes.Format_Character and token.value == FormatCharacters.Colon.value:
                if i+1 > tokens.count():
                    self._runtime.error(Error(ErrorCodesEnum.Expected, FormatCharacters.Colon.value))
                
                if i+2 > tokens.count():
                    self._runtime.error(Error(ErrorCodesEnum.Expected, Keywords.Type.value))

                if i+3 > tokens.count():
                    self._runtime.error(Error(ErrorCodesEnum.Expected, FormatCharacters.Left_Brace.value))

                ast.append(AST(ASTTypesEnum.Type, tokens[i+1].value, self._runtime.line_count, self._runtime.line_count))
                i += 3

            else:
                self._runtime.error(Error(ErrorCodesEnum.Unexpected, token.value))

            i += 1

        return AST(ASTTypesEnum.FuncDeclaration, ast, self._runtime.line_count, end)

    def _parse_use(self, tokens: List[Token]) -> AST:
        """ Parses use imports

        Args:
            tokens (List[Token]): Tokens from lexer

        AST:
            use Program;
            <use> <name> <end>
            use Program.Test;
            <use> <name> <.> <name> <end>

        Returns:
            AST: Library or class AST
        """
        ast = List(AST)
        i = 0
        # for i in range(0, tokens.count()):
        while i < tokens.count():
            token: Token = tokens[i]

            if i == tokens.count()-1 and token.type == TokenTypes.Format_Character and token.value == FormatCharacters.Semicolon.value:
                break

            elif i == tokens.count()-1:
                self._runtime.error(Error(ErrorCodesEnum.Expected, FormatCharacters.Semicolon.value))

            elif i == 0 and token.type == TokenTypes.Keyword and token.value == Keywords.Use.value:
                self._ast.append(AST(ASTTypesEnum.Keyword, token.value))

            elif i == 1 and token.type == TokenTypes.Name:
                Console.write_line('NAME')
                name, last_token = self._parse_name(tokens.skip(i))
                if last_token is not None:
                    i = tokens.index(last_token)
                    ast.append(name)

            else:
                Console.write_line('TEST')
                self._runtime.error(Error(ErrorCodesEnum.Unexpected, token.value))

            i += 1

        return AST(ASTTypesEnum.UseDirective, ast, self._runtime.line_count, self._runtime.line_count)

    def create_ast(self, tokens: List[Token]) -> List[AST]:
        self._ast = List(AST)

        if tokens.where(lambda t: t.type == TokenTypes.Keyword and t.value == Keywords.Use.value).count() > 0:
            self._ast.append(self._parse_use(tokens))

        elif tokens.where(lambda t: t.type == TokenTypes.Keyword and t.value == Keywords.Library.value).count() > 0:
            self._ast.append(self._parse_library_or_class(tokens))

        elif tokens.where(lambda t: t.type == TokenTypes.Keyword and t.value == Keywords.Class.value).count() > 0:
            self._ast.append(self._parse_library_or_class(tokens, True))

        elif tokens.where(lambda t: t.type == TokenTypes.Keyword and t.value == Keywords.Function.value).count() > 0:
            self._ast.append(self._parse_func(tokens))

        elif tokens.where(lambda t: t.type == TokenTypes.Keyword and t.value == Keywords.Variable.value).count() > 0:
            self._ast.append(self._parse_variable(tokens))

        return self._ast
